==> ppcboot-0.4.2.tar.gz: ppcboot-0.4.2/common/main.c <==
/*
 * (C) Copyright 2000
 * Wolfgang Denk, DENX Software Engineering, wd@denx.de.
 *
 * See file CREDITS for list of people who contributed to this
 * project.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 * MA 02111-1307 USA
 */

#include <ppcboot.h>
#include <command.h>

int readline (const char *const prompt);

static char * delete_char (char *buffer, char *p, int *colp, int *np, int plen);
static int parse_line (char *, char *[]);

char        console_buffer[CFG_CBSIZE];		/* console I/O buffer	*/

static char erase_seq[] = "\b \b";		/* erase sequence	*/
static char   tab_seq[] = "        ";		/* used to expand TABs	*/
static char   parse_buffer[CFG_CBSIZE];		/* cmd parse buffer	*/
static int   argc;
static char *argv[CFG_MAXARGS+1];		/* NULL terminated	*/

/****************************************************************************/

void main_loop(bd_t *bd)
{
	cmd_tbl_t *cmdtp;
#if (CONFIG_8xx_BOOTDELAY >= 0)
	char *s = getenv ("bootdelay");
	int bootdelay = s ? (int)simple_strtoul(s, NULL, 10) : 0;
	int autoboot  = 1;

#endif	/* CONFIG_8xx_BOOTDELAY */

#if 0
	printf ("### main_loop entered:\n\n");
#endif

	/* Set default arguments */
	argc = 0;
	argv[ argc ] =  NULL ;

	/*
	 * Main Loop for Monitor Command Processing
	 */
	
	for (;;) {
		int flag = 0;
		int len;

#if (CONFIG_8xx_BOOTDELAY >= 0)

		if (autoboot)
			printf ("Hit any key to stop autoboot: %2d ", bootdelay);

		while (bootdelay > 0) {
			int i;

			--bootdelay;
			/* delay 100 * 10ms */
			for (i=0; i<100; ++i) {
				if (serial_tstc()) {	/* we got a key press	*/
					bootdelay = 0;	/* no more delays	*/
					autoboot  = 0;	/* don't auto boot	*/
					(void) serial_getc();  /* consume input */
					break;
				}
				udelay (10000);
			}

			printf ("\b\b\b%2d ", bootdelay);
		}

		serial_putc ('\n');

		if (autoboot) {
			s = getenv("bootcmd");
			strncpy (console_buffer,
				 s ? s : "",
				 CFG_CBSIZE-1);
			console_buffer[CFG_CBSIZE-1] = '\0';	/* just in case */
			len = strlen (console_buffer);

			autoboot = 0;
		}
		else			/* No autoboot: read input		*/
#endif	/* CONFIG_8xx_BOOTDELAY */
		if ((len = readline (CFG_PROMPT)) < 0) {
			printf ("<INTERRUPT>\n");
			continue;
		}

		/*
		 * If we have any new input, we parse the new command line.
		 * Otherwise, we re-issue the previous command.
		 */
		if (len) {
			strcpy (parse_buffer, console_buffer);

			argc = parse_line (parse_buffer, argv);
		} else {
			flag |= CMD_FLAG_REPEAT;
		}
#if 0
	{	int i;
		printf ("ARGC = %d\n", argc);
		for (i=0; i<argc; ++i) {
			printf (">> ARGV[%d] = \"%s\"\n", i, argv[i]);
		}
		printf ("\n");
	}
#endif

		if (argc == 0) {	/* nothing to do */
			continue;
		}

		/*
		 * Search command table.
		 * Use linear search - it's a small table
		 */
		for (cmdtp=&cmd_tbl[0]; cmdtp->name; cmdtp++) {
			if (strncmp(argv[0], cmdtp->name, cmdtp->lmin) == 0) {
				/* found - check max args */
				if (argc > cmdtp->maxargs) {
					printf ("Usage:\n%s\n", cmdtp->usage);
					goto done;
				}
				/* OK - call function */
				(cmdtp->cmd)(cmdtp, bd, flag, argc, argv);
				goto done;
			}
		}
		printf ("Unknown command '%s' - try 'help'\n", argv[0]);
done:		;
	}
}

/****************************************************************************/

/*
 * Prompt for input and read a line.
 * Return number of read characters
 */
int readline (const char *const prompt)
{
	char   *p = console_buffer;
	int	n = 0;				/* buffer index		*/
	int	plen = strlen (prompt);		/* prompt length	*/
	int	col;				/* output column cnt	*/
	char	c;

	/* print prompt */
	if (prompt)
		serial_putstr (prompt);
	col = plen;

	for (;;) {
		c = serial_getc();

		/*
		 * Special character handling
		 */
		switch (c) {
		case '\r':				/* Enter		*/
		case '\n':
			*p = '\0';
			serial_putstr ("\r\n");
			return (p - console_buffer);
		
		case 0x03:				/* ^C - break		*/
			return (-1);
		
		case 0x15:				/* ^U - erase line	*/
			while (col > plen) {
				serial_putstr (erase_seq);
				--col;
			}
			p = console_buffer;
			n = 0;
			continue;

		case 0x17:				/* ^W - erase word 	*/
			p=delete_char(console_buffer, p, &col, &n, plen);
			while ((n > 0) && (*p != ' ')) {
				p=delete_char(console_buffer, p, &col, &n, plen);
			}
			continue;

		case 0x08:				/* ^H  - backspace	*/
		case 0x7F:				/* DEL - backspace	*/
			p=delete_char(console_buffer, p, &col, &n, plen);
			continue;

		default:
			/*
			 * Must be a normal character then
			 */
			if (n < CFG_CBSIZE-2) {
				if (c == '\t') {	/* expand TABs		*/
					serial_putstr (tab_seq+(col&07));
					col += 8 - (col&07);
				} else {
					++col;		/* echo input		*/
					serial_putc (c);
				}
				*p++ = c;
				++n;
			} else {			/* Buffer full		*/
				serial_putc ('\a');
			}
		}
	}
}

/****************************************************************************/

/*
 * Some commands are too dangerous to repeat when the user just presses
 * enter to see if the box is still alive.
 *
 * This function will turn off repeating of the current command.
 */

void	command_repeat_off (void)
{
	argc = 0;
}

/****************************************************************************/

static char * delete_char (char *buffer, char *p, int *colp, int *np, int plen)
{
	char *s;

	if (*np == 0) {
		return (p);
	}

	if (*(--p) == '\t') {			/* will retype the whole line	*/
		while (*colp > plen) {
			serial_putstr (erase_seq);
			(*colp)--;
		}
		for (s=buffer; s<p; ++s) {
			if (*s == '\t') {
				serial_putstr (tab_seq+((*colp) & 07));
				*colp += 8 - ((*colp) & 07);
			} else {
				++(*colp);
				serial_putc (*s);
			}
		}
	} else {
		serial_putstr (erase_seq);
		(*colp)--;
	}
	(*np)--;
	return (p);
}

/****************************************************************************/

int parse_line (char *line, char *argv[])
{
	int nargs = 0;

	while (nargs < CFG_MAXARGS) {

		/* skip any white space */
		while ((*line == ' ') || (*line == '\t')) {
			++line;
		}

		if (*line == '\0') {	/* end of line, no more args	*/
			argv[nargs] = NULL;
			return (nargs);
		}

		argv[nargs++] = line;	/* begin of argument string	*/

		/* find end of string */
		while (*line && (*line != ' ') && (*line != '\t')) {
			++line;
		}

		if (*line == '\0') {	/* end of line, no more args	*/
			argv[nargs] = NULL;
			return (nargs);
		}

		*line++ = '\0';		/* terminate current arg	 */
	}

	return (nargs);
}

/****************************************************************************/

==> ppcboot-0.5.1.tar.gz: ppcboot-0.5.1/common/main.c <==
/*
 * (C) Copyright 2000
 * Wolfgang Denk, DENX Software Engineering, wd@denx.de.
 *
 * See file CREDITS for list of people who contributed to this
 * project.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 * MA 02111-1307 USA
 */

#include <ppcboot.h>
#include <command.h>

int readline (const char *const prompt);

static char * delete_char (char *buffer, char *p, int *colp, int *np, int plen);
static int parse_line (char *, char *[]);

char        console_buffer[CFG_CBSIZE];		/* console I/O buffer	*/

static char erase_seq[] = "\b \b";		/* erase sequence	*/
static char   tab_seq[] = "        ";		/* used to expand TABs	*/
static char   parse_buffer[CFG_CBSIZE];		/* cmd parse buffer	*/
static int   argc;
static char *argv[CFG_MAXARGS+1];		/* NULL terminated	*/

/****************************************************************************/

void main_loop(bd_t *bd)
{
	cmd_tbl_t *cmdtp;
#if (CONFIG_BOOTDELAY >= 0)
	char *s = getenv ("bootdelay");
	int bootdelay = s ? (int)simple_strtoul(s, NULL, 10) : 0;
	int autoboot  = 1;

#endif	/* CONFIG_BOOTDELAY */

#if 0
	printf ("### main_loop entered:\n\n");
#endif

	/* Set default arguments */
	argc = 0;
	argv[ argc ] =  NULL ;

	/*
	 * Main Loop for Monitor Command Processing
	 */
	
	for (;;) {
		int flag = 0;
		int len;

#if (CONFIG_BOOTDELAY >= 0)

		if (autoboot)
			printf ("Hit any key to stop autoboot: %2d ", bootdelay);

		while (bootdelay > 0) {
			int i;

			--bootdelay;
			/* delay 100 * 10ms */
			for (i=0; i<100; ++i) {
				if (serial_tstc()) {	/* we got a key press	*/
					bootdelay = 0;	/* no more delays	*/
					autoboot  = 0;	/* don't auto boot	*/
					(void) serial_getc();  /* consume input */
					break;
				}
				udelay (10000);
			}

			printf ("\b\b\b%2d ", bootdelay);
		}

		serial_putc ('\n');

		if (autoboot) {
			s = getenv("bootcmd");
			strncpy (console_buffer,
				 s ? s : "",
				 CFG_CBSIZE-1);
			console_buffer[CFG_CBSIZE-1] = '\0';	/* just in case */
			len = strlen (console_buffer);

			autoboot = 0;
		}
		else			/* No autoboot: read input		*/
#endif	/* CONFIG_BOOTDELAY */
		if ((len = readline (CFG_PROMPT)) < 0) {
			printf ("<INTERRUPT>\n");
			continue;
		}

		/*
		 * If we have any new input, we parse the new command line.
		 * Otherwise, we re-issue the previous command.
		 */
		if (len) {
			strcpy (parse_buffer, console_buffer);

			argc = parse_line (parse_buffer, argv);
		} else {
			flag |= CMD_FLAG_REPEAT;
		}
#if 0
	{	int i;
		printf ("ARGC = %d\n", argc);
		for (i=0; i<argc; ++i) {
			printf (">> ARGV[%d] = \"%s\"\n", i, argv[i]);
		}
		printf ("\n");
	}
#endif

		if (argc == 0) {	/* nothing to do */
			continue;
		}

		/*
		 * Search command table.
		 * Use linear search - it's a small table
		 */
		for (cmdtp=&cmd_tbl[0]; cmdtp->name; cmdtp++) {
			if (strncmp(argv[0], cmdtp->name, cmdtp->lmin) == 0) {
				/* found - check max args */
				if (argc > cmdtp->maxargs) {
					printf ("Usage:\n%s\n", cmdtp->usage);
					goto done;
				}
				/* OK - call function */
				(cmdtp->cmd)(cmdtp, bd, flag, argc, argv);
				goto done;
			}
		}
		printf ("Unknown command '%s' - try 'help'\n", argv[0]);
done:		;
	}
}

/****************************************************************************/

/*
 * Prompt for input and read a line.
 * Return number of read characters
 */
int readline (const char *const prompt)
{
	char   *p = console_buffer;
	int	n = 0;				/* buffer index		*/
	int	plen = strlen (prompt);		/* prompt length	*/
	int	col;				/* output column cnt	*/
	char	c;

	/* print prompt */
	if (prompt)
		serial_putstr (prompt);
	col = plen;

	for (;;) {
		c = serial_getc();

		/*
		 * Special character handling
		 */
		switch (c) {
		case '\r':				/* Enter		*/
		case '\n':
			*p = '\0';
			serial_putstr ("\r\n");
			return (p - console_buffer);
		
		case 0x03:				/* ^C - break		*/
			return (-1);
		
		case 0x15:				/* ^U - erase line	*/
			while (col > plen) {
				serial_putstr (erase_seq);
				--col;
			}
			p = console_buffer;
			n = 0;
			continue;

		case 0x17:				/* ^W - erase word 	*/
			p=delete_char(console_buffer, p, &col, &n, plen);
			while ((n > 0) && (*p != ' ')) {
				p=delete_char(console_buffer, p, &col, &n, plen);
			}
			continue;

		case 0x08:				/* ^H  - backspace	*/
		case 0x7F:				/* DEL - backspace	*/
			p=delete_char(console_buffer, p, &col, &n, plen);
			continue;

		default:
			/*
			 * Must be a normal character then
			 */
			if (n < CFG_CBSIZE-2) {
				if (c == '\t') {	/* expand TABs		*/
					serial_putstr (tab_seq+(col&07));
					col += 8 - (col&07);
				} else {
					++col;		/* echo input		*/
					serial_putc (c);
				}
				*p++ = c;
				++n;
			} else {			/* Buffer full		*/
				serial_putc ('\a');
			}
		}
	}
}

/****************************************************************************/

/*
 * Some commands are too dangerous to repeat when the user just presses
 * enter to see if the box is still alive.
 *
 * This function will turn off repeating of the current command.
 */

void	command_repeat_off (void)
{
	argc = 0;
}

/****************************************************************************/

static char * delete_char (char *buffer, char *p, int *colp, int *np, int plen)
{
	char *s;

	if (*np == 0) {
		return (p);
	}

	if (*(--p) == '\t') {			/* will retype the whole line	*/
		while (*colp > plen) {
			serial_putstr (erase_seq);
			(*colp)--;
		}
		for (s=buffer; s<p; ++s) {
			if (*s == '\t') {
				serial_putstr (tab_seq+((*colp) & 07));
				*colp += 8 - ((*colp) & 07);
			} else {
				++(*colp);
				serial_putc (*s);
			}
		}
	} else {
		serial_putstr (erase_seq);
		(*colp)--;
	}
	(*np)--;
	return (p);
}

/****************************************************************************/

int parse_line (char *line, char *argv[])
{
	int nargs = 0;

	while (nargs < CFG_MAXARGS) {

		/* skip any white space */
		while ((*line == ' ') || (*line == '\t')) {
			++line;
		}

		if (*line == '\0') {	/* end of line, no more args	*/
			argv[nargs] = NULL;
			return (nargs);
		}

		argv[nargs++] = line;	/* begin of argument string	*/

		/* find end of string */
		while (*line && (*line != ' ') && (*line != '\t')) {
			++line;
		}

		if (*line == '\0') {	/* end of line, no more args	*/
			argv[nargs] = NULL;
			return (nargs);
		}

		*line++ = '\0';		/* terminate current arg	 */
	}

	return (nargs);
}

/****************************************************************************/
